@@ -70,6 +70,7 @@ module Aws.Core
 , IteratedTransaction(..)
   -- * Credentials
 , Credentials(..)
+, makeCredentials
 , credentialsDefaultFile
 , credentialsDefaultKey
 , loadCredentialsFromFile
@@ -121,6 +122,7 @@ import qualified Data.Text                as T
 import qualified Data.Text.Encoding       as T
 import qualified Data.Text.IO             as T
 import           Data.Time
+import qualified Data.Traversable         as Traversable
 import           Data.Typeable
 import           Data.Word
 import qualified Network.HTTP.Conduit     as HTTP
@@ -128,16 +130,22 @@ import qualified Network.HTTP.Types       as HTTP
 import           System.Directory
 import           System.Environment
 import           System.FilePath          ((</>))
+#if MIN_VERSION_time(1,5,0)
+import           Data.Time.Format
+#else
 import           System.Locale
+#endif
 import qualified Text.XML                 as XML
 import qualified Text.XML.Cursor          as Cu
 import           Text.XML.Cursor          hiding (force, forceM)
+-------------------------------------------------------------------------------
 
 -- | Types that can be logged (textually).
 class Loggable a where
     toLogText :: a -> T.Text
 
--- | A response with metadata. Can also contain an error response, or an internal error, via 'Attempt'.
+-- | A response with metadata. Can also contain an error response, or
+-- an internal error, via 'Attempt'.
 --
 -- Response forms a Writer-like monad.
 data Response m a = Response { responseMetadata :: m
@@ -189,10 +197,12 @@ type HTTPResponseConsumer a = HTTP.Response (C.ResumableSource (ResourceT IO) By
 --
 -- Note that for debugging, there is an instance for 'L.ByteString'.
 class Monoid (ResponseMetadata resp) => ResponseConsumer req resp where
-    -- | Metadata associated with a response. Typically there is one metadata type for each AWS service.
+    -- | Metadata associated with a response. Typically there is one
+    -- metadata type for each AWS service.
     type ResponseMetadata resp
 
-    -- | Response parser. Takes the corresponding request, an 'IORef' for metadata, and HTTP response data.
+    -- | Response parser. Takes the corresponding request, an 'IORef'
+    -- for metadata, and HTTP response data.
     responseConsumer :: req -> IORef (ResponseMetadata resp) -> HTTPResponseConsumer resp
 
 -- | Does not parse response. For debugging.
@@ -213,18 +223,20 @@ class AsMemoryResponse resp where
 class ListResponse resp item | resp -> item where
     listResponse :: resp -> [item]
 
+
 -- | Associates a request type and a response type in a bi-directional way.
 --
--- This allows the type-checker to infer the response type when given the request type and vice versa.
+-- This allows the type-checker to infer the response type when given
+-- the request type and vice versa.
 --
--- Note that the actual request generation and response parsing resides in 'SignQuery' and 'ResponseConsumer'
--- respectively.
+-- Note that the actual request generation and response parsing
+-- resides in 'SignQuery' and 'ResponseConsumer' respectively.
 class (SignQuery r, ResponseConsumer r a, Loggable (ResponseMetadata a))
       => Transaction r a
-      | r -> a, a -> r
+      | r -> a
 
 -- | A transaction that may need to be split over multiple requests, for example because of upstream response size limits.
-class Transaction r a => IteratedTransaction r a | r -> a , a -> r where
+class Transaction r a => IteratedTransaction r a | r -> a where
     nextIteratedRequest :: r -> a -> Maybe r
 
 -- | Signature version 4: ((region, service),(date,key))
@@ -245,6 +257,15 @@ data Credentials
 instance Show Credentials where
     show c = "Credentials{accessKeyID=" ++ show (accessKeyID c) ++ ",secretAccessKey=" ++ show (secretAccessKey c) ++ ",iamToken=" ++ show (iamToken c) ++ "}"
 
+makeCredentials :: MonadIO io
+                => B.ByteString -- ^ AWS Access Key ID
+                -> B.ByteString -- ^ AWS Secret Access Key
+                -> io Credentials
+makeCredentials accessKeyID secretAccessKey = liftIO $ do
+    v4SigningKeys <- newIORef []
+    let iamToken = Nothing
+    return Credentials { .. }
+
 -- | The file where access credentials are loaded, when using 'loadCredentialsDefault'.
 --
 -- Value: /<user directory>/@/.aws-keys@
@@ -264,32 +285,29 @@ credentialsDefaultKey = "default"
 -- @keyName awsKeyID awsKeySecret@
 loadCredentialsFromFile :: MonadIO io => FilePath -> T.Text -> io (Maybe Credentials)
 loadCredentialsFromFile file key = liftIO $ do
-  contents <- map T.words . T.lines <$> T.readFile file
-  ref <- newIORef []
-  return $ do
-    [_key, keyID, secret] <- find (hasKey key) contents
-    return Credentials { accessKeyID = T.encodeUtf8 keyID
-                       , secretAccessKey = T.encodeUtf8 secret
-                       , v4SigningKeys = ref
-                       , iamToken = Nothing
-                       }
-      where
-        hasKey _ [] = False
-        hasKey k (k2 : _) = k == k2
+  exists <- doesFileExist file
+  if exists
+    then do
+      contents <- map T.words . T.lines <$> T.readFile file
+      Traversable.sequence $ do
+        [_key, keyID, secret] <- find (hasKey key) contents
+        return (makeCredentials (T.encodeUtf8 keyID) (T.encodeUtf8 secret))
+    else return Nothing
+  where
+    hasKey _ [] = False
+    hasKey k (k2 : _) = k == k2
 
 -- | Load credentials from the environment variables @AWS_ACCESS_KEY_ID@ and @AWS_ACCESS_KEY_SECRET@
 --   (or @AWS_SECRET_ACCESS_KEY@), if possible.
 loadCredentialsFromEnv :: MonadIO io => io (Maybe Credentials)
 loadCredentialsFromEnv = liftIO $ do
   env <- getEnvironment
-  ref <- newIORef []
   let lk = flip lookup env
       keyID = lk "AWS_ACCESS_KEY_ID"
       secret = lk "AWS_ACCESS_KEY_SECRET" `mplus` lk "AWS_SECRET_ACCESS_KEY"
-  return (Credentials <$> (T.encodeUtf8 . T.pack <$> keyID) 
-                      <*> (T.encodeUtf8 . T.pack <$> secret)
-                      <*> return ref
-                      <*> return Nothing)
+  Traversable.sequence
+      (makeCredentials <$> (T.encodeUtf8 . T.pack <$> keyID)
+                       <*> (T.encodeUtf8 . T.pack <$> secret))
 
 loadCredentialsFromInstanceMetadata :: MonadIO io => io (Maybe Credentials)
 loadCredentialsFromInstanceMetadata = liftIO $ HTTP.withManager $ \mgr ->
@@ -341,10 +359,10 @@ loadCredentialsFromEnvOrFileOrInstanceMetadata file key =
       Just cr -> return (Just cr)
       Nothing ->
         do
-          instcr <- loadCredentialsFromInstanceMetadata
-          case instcr of
+          filecr <- loadCredentialsFromFile file key
+          case filecr of
             Just cr -> return (Just cr)
-            Nothing -> loadCredentialsFromFile file key
+            Nothing -> loadCredentialsFromInstanceMetadata
 
 -- | Load credentials from environment variables if possible, or alternative from the default file with the default
 -- key name.
@@ -362,7 +380,7 @@ loadCredentialsDefault = do
 data Protocol
     = HTTP
     | HTTPS
-    deriving (Show)
+    deriving (Eq,Read,Show,Ord,Typeable)
 
 -- | The default port to be used for a protocol if no specific port is specified.
 defaultPort :: Protocol -> Int
@@ -378,7 +396,7 @@ data Method
     | Post      -- ^ POST method. Sends a service- and request-specific request body.
     | Put       -- ^ PUT method.
     | Delete    -- ^ DELETE method.
-    deriving (Show, Eq)
+    deriving (Show, Eq, Ord)
 
 -- | HTTP method associated with a request method.
 httpMethod :: Method -> HTTP.Method
@@ -393,37 +411,37 @@ httpMethod Delete    = "DELETE"
 data SignedQuery
     = SignedQuery {
         -- | Request method.
-        sqMethod :: Method
+        sqMethod :: !Method
         -- | Protocol to be used.
-      , sqProtocol :: Protocol
+      , sqProtocol :: !Protocol
         -- | HTTP host.
-      , sqHost :: B.ByteString
+      , sqHost :: !B.ByteString
         -- | IP port.
-      , sqPort :: Int
+      , sqPort :: !Int
         -- | HTTP path.
-      , sqPath :: B.ByteString
+      , sqPath :: !B.ByteString
         -- | Query string list (used with 'Get' and 'PostQuery').
-      , sqQuery :: HTTP.Query
+      , sqQuery :: !HTTP.Query
         -- | Request date/time.
-      , sqDate :: Maybe UTCTime
+      , sqDate :: !(Maybe UTCTime)
         -- | Authorization string (if applicable), for @Authorization@ header.  See 'authorizationV4'
-      , sqAuthorization :: Maybe (IO B.ByteString)
+      , sqAuthorization :: !(Maybe (IO B.ByteString))
         -- | Request body content type.
-      , sqContentType :: Maybe B.ByteString
+      , sqContentType :: !(Maybe B.ByteString)
         -- | Request body content MD5.
-      , sqContentMd5 :: Maybe (Digest MD5)
+      , sqContentMd5 :: !(Maybe (Digest MD5))
         -- | Additional Amazon "amz" headers.
-      , sqAmzHeaders :: HTTP.RequestHeaders
+      , sqAmzHeaders :: !HTTP.RequestHeaders
         -- | Additional non-"amz" headers.
-      , sqOtherHeaders :: HTTP.RequestHeaders
+      , sqOtherHeaders :: !HTTP.RequestHeaders
         -- | Request body (used with 'Post' and 'Put').
 #if MIN_VERSION_http_conduit(2, 0, 0)
-      , sqBody :: Maybe HTTP.RequestBody
+      , sqBody :: !(Maybe HTTP.RequestBody)
 #else
-      , sqBody :: Maybe (HTTP.RequestBody (C.ResourceT IO))
+      , sqBody :: !(Maybe (HTTP.RequestBody (C.ResourceT IO)))
 #endif
         -- | String to sign. Note that the string is already signed, this is passed mostly for debugging purposes.
-      , sqStringToSign :: B.ByteString
+      , sqStringToSign :: !B.ByteString
       }
     --deriving (Show)
 
@@ -443,25 +461,39 @@ queryToHttpRequest SignedQuery{..} =  do
       , HTTP.host = sqHost
       , HTTP.port = sqPort
       , HTTP.path = sqPath
-      , HTTP.queryString = HTTP.renderQuery False sqQuery
+      , HTTP.queryString =
+          if sqMethod == PostQuery
+            then ""
+            else HTTP.renderQuery False sqQuery
+
       , HTTP.requestHeaders = catMaybes [ checkDate (\d -> ("Date", fmtRfc822Time d)) sqDate
                                         , fmap (\c -> ("Content-Type", c)) contentType
                                         , fmap (\md5 -> ("Content-MD5", Base64.encode $ toBytes md5)) sqContentMd5
                                         , fmap (\auth -> ("Authorization", auth)) mauth]
                               ++ sqAmzHeaders
                               ++ sqOtherHeaders
-      , HTTP.requestBody = case sqMethod of
-                             PostQuery -> HTTP.RequestBodyLBS . Blaze.toLazyByteString $ HTTP.renderQueryBuilder False sqQuery
-                             _         -> case sqBody of
-                                            Nothing -> HTTP.RequestBodyBuilder 0 mempty
-                                            Just x  -> x
+      , HTTP.requestBody =
+
+        -- An explicityly defined body parameter should overwrite everything else.
+        case sqBody of
+          Just x -> x
+          Nothing ->
+            -- a POST query should convert its query string into the body
+            case sqMethod of
+              PostQuery -> HTTP.RequestBodyLBS . Blaze.toLazyByteString $
+                           HTTP.renderQueryBuilder False sqQuery
+              _         -> HTTP.RequestBodyBuilder 0 mempty
+
       , HTTP.decompress = HTTP.alwaysDecompress
       , HTTP.checkStatus = \_ _ _ -> Nothing
       }
-    where contentType = case sqMethod of
-                           PostQuery -> Just "application/x-www-form-urlencoded; charset=utf-8"
-                           _ -> sqContentType
-          checkDate f mb = maybe (f <$> mb) (const Nothing) $ lookup "date" sqOtherHeaders
+    where
+      checkDate f mb = maybe (f <$> mb) (const Nothing) $ lookup "date" sqOtherHeaders
+      -- An explicitly defined content-type should override everything else.
+      contentType = sqContentType `mplus` defContentType
+      defContentType = case sqMethod of
+                         PostQuery -> Just "application/x-www-form-urlencoded; charset=utf-8"
+                         _ -> Nothing
 
 -- | Create a URI fro a 'SignedQuery' object.
 --
@@ -562,7 +594,7 @@ signature cr ah input = Base64.encode sig
 -- See <http://docs.aws.amazon.com/general/latest/gr/signature-version-4.html>: you must create the
 -- canonical request as explained by Step 1 and this function takes care of Steps 2 and 3.
 authorizationV4 :: SignatureData
-                -> AuthorizationHash 
+                -> AuthorizationHash
                 -> B.ByteString -- ^ region, e.g. us-east-1
                 -> B.ByteString -- ^ service, e.g. dynamodb
                 -> B.ByteString -- ^ SignedHeaders, e.g. content-type;host;x-amz-date;x-amz-target
@@ -643,8 +675,9 @@ class DefaultServiceConfiguration config where
     debugServiceConfig :: config
     debugServiceConfig = defServiceConfig
 
--- | @queryList f prefix xs@ constructs a query list from a list of elements @xs@, using a common prefix @prefix@,
--- and a transformer function @f@.
+-- | @queryList f prefix xs@ constructs a query list from a list of
+-- elements @xs@, using a common prefix @prefix@, and a transformer
+-- function @f@.
 --
 -- A dot (@.@) is interspersed between prefix and generated key.
 --
@@ -743,7 +776,7 @@ newtype FormException = FormException { formErrorMesage :: String }
 instance E.Exception FormException
 
 -- | No credentials were found and an invariant was violated.
-newtype NoCredentialsException = NoCredentialsException { noCredentialsErrorMesage :: String }
+newtype NoCredentialsException = NoCredentialsException { noCredentialsErrorMessage :: String }
     deriving (Show, Typeable)
 
 instance E.Exception NoCredentialsException
